#include "logger.h"

int pyfcgi_logger_init()
{
	pyfcgi_conf_logger_t *conf;
	conf = &PyFCGI_conf.logs;
	memset(conf, 0, sizeof(pyfcgi_conf_logger_t));
	return 0;
}

int pyfcgi_logger_add(const char *filename, logmask_t loglvl, logmask_t logtyp,
                      const char *format)
{
	pyfcgi_conf_logger_t *conf;
	pyfcgi_logger_t logger;
	size_t new_idx;
	void *tmp;
	int err;
	char *err_fmt;

	conf = &PyFCGI_conf.logs;

	if(conf->logger_sz == 255)
	{
		pyfcgi_log(LOG_ERR, "Maximum of 255 logger reached, unable to add this one...");
		return PYFCGI_ERR;
	}

	if(pyfcgi_logger_format_add(format, &(logger.fmt_id)))
	{
		return PYFCGI_FATAL;
	}
	if(!logger.filename)
	{
		err = errno;
		err_fmt = "Unable to duplicate logger filename : %s";
		goto err;
	}

	logger.loglvl = loglvl;
	logger.logtyp = logtyp;
	logger.filename = strdup(filename);

	new_idx = conf->logger_sz;
	conf->logger_sz++;
	if( !(tmp = realloc(conf->loggers,
	                    sizeof(pyfcgi_logger_t) * conf->logger_sz)) )
	{
		err = errno;
		err_fmt = "Unable to realloc loggers array : %s";
		goto err_free_filename;
	}
	conf->loggers[new_idx] = logger;
	return 0;

err_free_filename:
	free(logger.filename);
err:
	pyfcgi_log(LOG_ALERT,
	           err_fmt,
		   strerror(err));
	return PYFCGI_FATAL;	
}

int pyfcgi_logger_format_add(const char* format, size_t* idx)
{
	char *fmt;
	size_t i;
	void *tmp;
	pyfcgi_conf_logger_t *conf;

	conf = &PyFCGI_conf.logs;

	for(i=0; i<conf->format_sz; i++)
	{
		if(!strcmp(format, conf->format[i])) { break; }
	}
	if(idx) { *idx = i; }
	if( i<conf->format_sz )
	{
		return 0;
	}
	if( !(fmt = strdup(format)) )
	{
		pyfcgi_log(LOG_ALERT, "Fails to strdup new format : %s",
			   strerror(errno));
		return PYFCGI_FATAL;
	}
	conf->format_sz++;
	if( !(tmp = realloc(conf->format,
	                    sizeof(char*) * conf->format_sz)) )
	{
		conf->format_sz--;
		pyfcgi_log(LOG_ALERT, "Fails to realloc logger format array : %s",
			   strerror(errno));
		return PYFCGI_FATAL;
	}
	conf->format[i] = fmt;
	return 0;
}

int pyfcgi_logger_parse_format(const char* fmt,
                               pyfcgi_logger_format_t* fmt_data)
{
	const char *ptr, *str_start;
	int err;
	char reason_err[PYFCGI_LOGGER_FMT_PARSE_ERRSZ];
	unsigned char nfield;
	pyfcgi_logger_fmt_field_t *cur_field;

	memset(fmt_data, 0, sizeof(pyfcgi_logger_format_t));
	ptr = fmt;
	nfield = 0;
	cur_field = fmt_data->fields;

	while(*ptr)
	{
		if(nfield >= PYFCGI_LOGGER_FIELD_MAX)
		{
			//TODO error truncate !
			break;
		}
		str_start = ptr;
		while(*ptr && *ptr != '{' && *(ptr+1) != '{')
		{
			if(!cur_field->type)
			{
				cur_field->type = pyfcgi_logger_field_const;
				cur_field->known_length = 1;
				cur_field->len = 0;
			}
			if(*ptr == '{' && *(ptr+1) == '{')
			{
				cur_field->len++;
				ptr++;
			}
			cur_field->len++;
			ptr++;
		}
		if(cur_field->len)
		{
			cur_field->val = strndup(str_start, cur_field->len);
			if(!cur_field->val)
			{
				err = errno;
				pyfcgi_log(LOG_ALERT,
				           "Unable to strdup part of format : %s",
					   strerror(err));
				goto exit_err;
			}
			nfield++;
			cur_field = &(fmt_data->fields[nfield]); // next_field
			continue; // to check nfield
		}
		if(*ptr == '{') //should be this or '\0'...
		{
			ptr++;
			if( (1 == pyfcgi_logger_parse_field(&ptr, fmt, 
			                                    cur_field, 
			                                    reason_err)) )
			{
				goto parse_err;
			}
		}
		ptr++;
	};
	

	return 0;

parse_err:
	pyfcgi_log(LOG_ERR, "Logger format parse error at chr %ld : %s",
	           ptr - fmt, reason_err);
exit_err:
	//TODO free const str from fields
	memset(fmt_data, 0, sizeof(pyfcgi_logger_format_t));
	return PYFCGI_ERR;
}

int pyfcgi_logger_parse_field(const char** ptr, const char *start,
                              pyfcgi_logger_fmt_field_t* cur_field,
			      char fail_reason[PYFCGI_LOGGER_FMT_PARSE_ERRSZ])
{
	
	const char *str_start, *field_start;
	int ret;
	char **name, **format;
	size_t default_len;
	pyfcgi_logger_field_type_e *type;
	char *field_names[] = {"datetime", "level", "facility", "pid", "ident",
	                       "msg", NULL};
	pyfcgi_logger_field_type_e field_types[] = {
		pyfcgi_logger_field_datetime, pyfcgi_logger_field_level,
		pyfcgi_logger_field_facility, pyfcgi_logger_field_pid,
		pyfcgi_logger_field_ident, pyfcgi_logger_field_msg};

	str_start = *ptr;
	field_start = *ptr;
	name = field_names;
	type = field_types;
	while(*name)
	{
		if(**name != **ptr)
		{
			name++;
			type++;
			continue;
		}
		// Name found ?
		while(**ptr && **name && **ptr != ':')
		{
			if(**ptr != **name)
			{
				break;
			}
			(*ptr)++;
			(*name)++;
		}
		if(**ptr == ':' || **ptr == '}')
		{
			cur_field->type = *type;
		}
		else
		{
			snprintf(fail_reason, PYFCGI_LOGGER_FMT_PARSE_ERRSZ,
				 "unknown field type '%s' chr %ld",
				 str_start, str_start - start);
			return 1;
		}
	}
	(*ptr)++; //on field size or ':' or '}'
	str_start = *ptr;
	if(pyfcgi_logger_parse_field_sz(ptr, &(cur_field->len)))
	{
		snprintf(fail_reason, PYFCGI_LOGGER_FMT_PARSE_ERRSZ,
			 "Unable to parse field size chr %ld '%s'",
			 str_start - start, str_start);
		return 1;
	}
	if(**ptr != ':' && **ptr != '}')
	{
		snprintf(fail_reason, PYFCGI_LOGGER_FMT_PARSE_ERRSZ,
			 "Unable to parse field '%s' at chr %ld",
			 field_start , *ptr - start);
		return 1;
	}
	switch(cur_field->type)
	{
		case pyfcgi_logger_field_datetime:
			default_len = 10;
			cur_field->val = (void*)strftime;
			format = &(cur_field->args.datetime.format);
			ret = pyfcgi_logger_parse_field_dtfmt(ptr,format);
			if(ret){
				return ret;
			}
			break;
		case pyfcgi_logger_field_level:
			default_len = 7;
			break;
		case pyfcgi_logger_field_facility:
			default_len = 6;
			break;
		case pyfcgi_logger_field_pid:
			default_len = 5;
			break;
		case pyfcgi_logger_field_ident:
			default_len = 0;
			break;
		case pyfcgi_logger_field_msg:
			default_len = 0;
			break;
		default:
			snprintf(fail_reason, PYFCGI_LOGGER_FMT_PARSE_ERRSZ,
				 "Unknown error parsing field '%s' at chr %ld",
				 field_start , *ptr - start);
			return 1;
	}
	if(!cur_field->len && default_len)
	{
		cur_field->len = default_len;
	}
	return 0;
}

int pyfcgi_logger_parse_field_dtfmt(const char** ptr, char** format)
{
	const char *fmt;
	size_t fmt_len;

	fmt_len = 0;
	fmt = *ptr;
printf("FUCKOFF");
	while(**ptr && **ptr != '}' && *((*ptr)+1) != '}')
	{
		ptr++;
		fmt_len++;
	}
	if(!fmt_len)
	{
		fmt_len = sizeof(PYFCGI_LOGGER_TIME_FMT_DEFAULT);
		fmt = PYFCGI_LOGGER_TIME_FMT_DEFAULT;
	}
	if(!(*format = strndup(fmt, fmt_len)))
	{
		pyfcgi_log(LOG_ALERT,
		           "Unable to strdup field datetime format : %s",
			   strerror(errno));
		return PYFCGI_FATAL;
	}
	return 0;
}

int pyfcgi_logger_parse_field_sz(const char **ptr, size_t *size)
{
	char *end;
	if(**ptr == '}' || **ptr == ':')
	{
		return (*size = 0);
	}
	if(**ptr < '0' && **ptr > 9)
	{
		return 1; // parse error
	}
	*size = strtoull(*ptr, &end, 10);
	return 0;
}

int pyfcgi_logger_open(pyfcgi_logger_t *logger)
{
	if( (logger->fd = open(logger->filename,
	                       O_WRONLY | O_APPEND | O_CREAT, 00640) < 0) )
	{
		pyfcgi_log(LOG_ERR,
		           "Unable to open log file '%s' : %s",
			   strerror(errno));
		return PYFCGI_ERR;
	}
	return 0;
}

int pyfcgi_log(loglvl_t lvl, const char *fmt, ...)
{
	int ret;
	va_list ap;
	va_start(ap, fmt);
	ret = vpyfcgi_log(lvl, fmt, ap);
	va_end(ap);
	return ret;
}

int vpyfcgi_log(loglvl_t lvl, const char *fmt, va_list ap)
{
	int len, ret;
	pyfcgi_conf_logger_t *conf;
	unsigned char i;
	char buf[512];

	conf = &PyFCGI_conf.logs;

	_vsyslog(lvl, fmt, ap);

	len = 0;
	if(conf->format_sz > 1)
	{
		len = vdprintf(conf->pipes[1], fmt, ap);
		if(len < 0)
		{
			_syslog(LOG_ALERT, 
			       "Unable to write to multiplexer pipe when trying to log : %s",
			       strerror(errno));
			return PYFCGI_FATAL;
		}
	}
	else if (conf->format_sz)
	{
		len = vdprintf(conf->loggers[0].fd, fmt, ap);
		if(len < 0)
		{
			_syslog(LOG_ALERT, 
			       "Unable to write to single FD to '%s' when trying to log : %s",
			       conf->loggers[0].filename, strerror(errno));
			return PYFCGI_FATAL;
		}
		return 0;
	}
	for(i=0; i<conf->logger_sz-1; i++)
	{
		ret = tee(conf->pipes[0], conf->loggers[i].fd, len, 0);
		if(ret < 0)
		{
			_syslog(LOG_ALERT, 
			       "Unable to splice to last logfile '%s' : %s",
			       conf->loggers[i].filename, strerror(errno));
			return PYFCGI_FATAL;
		}
	}
	ret = splice(conf->pipes[0], NULL, conf->loggers[i].fd, NULL, len, 0);
	if(ret < 0)
	{
		_syslog(LOG_ALERT, 
		       "Unable to splice to last logfile '%s' : %s",
		       conf->loggers[i].filename, strerror(errno));
		return PYFCGI_ERR;
	}
	if(ret < len)
	{
		_syslog(LOG_WARNING, 
		       "Unable to splice all data to logfile. Flushing pipe.");

		do
		{
			len -= (ret<len)?ret:len;
			ret = read(conf->pipes[0], buf, (len>sizeof(buf))?sizeof(buf):len);
			if(ret < 0)
			{
				_syslog(LOG_CRIT,
				        "Logger pipe seems broken, unable to flush : %s",
					strerror(errno));
				return PYFCGI_ERR;
			}
		}while(len);
	}
	return 0;
}
